--CPML is the magic that makes the transforms work
local cpml = require "cpml"
local modelLoader = require "modelLoader"

local gun = {}
local egg = {}
local cube = {}
local shader, matrix, makeProper, drawModel

--initial position and angle of Camera
local cameraPos = cpml.vec3(0, 0, -15)
local angle     = cpml.vec2(0, 0)

--initial transforms for the model
local gunTransform = cpml.mat4.identity()

local eggTransform = cpml.mat4.identity()
eggTransform:translate(eggTransform, cpml.vec3(12,0,0))

local cubeTransform = cpml.mat4.identity()
cubeTransform:translate(cubeTransform, cpml.vec3(-12,0,0))


--Custom vertex format add z, w to vertex position and the normals for vertices
local format = {
	{
		"VertexPosition",
		"float",
		4
	},
	{
		"VertexTexCoord",
		"float",
		2
	},
	{
		"VertexNormal",
		"float",
		3
	}
}


function love.load()
	love.graphics.setBackgroundColor(.5,.5,.5)

	local gunMesh = love.graphics.newMesh(format, modelLoader("/models/glock.obj"), "triangles")
	gunMesh:setTexture(love.graphics.newImage("/images/carbon.png"))

	gun = {
		transform = gunTransform,
		mesh = gunMesh
	}

	local eggMesh = love.graphics.newMesh(format, modelLoader("/models/egg.obj"), "triangles")
	eggMesh:setTexture(love.graphics.newImage("/images/egg.png"))

	egg = {
		transform = eggTransform,
		mesh = eggMesh
	}

	local cubeMesh = love.graphics.newMesh(format, modelLoader("/models/cube.obj"), "triangles")
	cubeMesh:setTexture(love.graphics.newImage("/images/boxtexture.png"))

	cube = {
		transform = cubeTransform,
		mesh = cubeMesh
	}

	shader = love.graphics.newShader("basicRender.glsl")

	--Sets DepthMode. Prevents items being drawn out of order
	love.graphics.setDepthMode("lequal", true)

end

function love.draw()

	drawModel(gun)
	drawModel(egg)
	drawModel(cube)

	love.graphics.print("Angle: " .. angle.x .. ", " .. angle.y, 20, 20)
	love.graphics.print("Position: " .. cameraPos.x .. ", " .. cameraPos.y, 20, 20*2)
	love.graphics.print("Backspace to reset position and angle", 20, 20*3)
	love.graphics.print("Click and drag to rotate", 20, 20*4)
	love.graphics.print("WASD to move camera position", 20, 20*5)
	love.graphics.print("FPS " .. love.timer.getFPS(), 20, 20*6)
end

function love.update(dt)
	local speed = 5 * dt
	if love.keyboard.isDown("w", "up") then
		cameraPos.y = cameraPos.y - speed
	elseif love.keyboard.isDown("s", "down") then
		cameraPos.y = cameraPos.y + speed
	end
	if love.keyboard.isDown("a", "left") then
		cameraPos.x = cameraPos.x + speed
	elseif love.keyboard.isDown("d", "right") then
		cameraPos.x = cameraPos.x - speed
	end
end

--Function to merge camera transform and projection matrix(I think anyways)
function matrix()
	local mat = cpml.mat4.from_perspective(60, love.graphics.getWidth()/love.graphics.getHeight(), 0.1, 1000)
	local v = cpml.mat4()
	v:translate(v, cameraPos)
	v:rotate(v, -math.pi/2, cpml.vec3.unit_x)
	v:rotate(v, angle.y, cpml.vec3.unit_x)
	v:rotate(v, angle.x, cpml.vec3.unit_y)
	return makeProper(mat) * makeProper(v)
end


--Function to correct cpml incorrect matrix format.
function makeProper(mat)
	local m = {}
	for x = 1, 4 do
		local t = {}
		for y = 1, 4 do
			table.insert(t, mat[(y - 1) * 4 + x])
		end
		table.insert(m, t)
	end
	return cpml.mat4(m)
end

function love.keypressed(key)
	if key == "escape" then
		love.event.quit()
	elseif key == "backspace" then
		angle = cpml.vec2(0, 0)
		cameraPos = cpml.vec3(0, 0, -15)
	end
end

function love.mousepressed(x, y, button)
	if button == 1 then
		dragging = true
		love.mouse.setRelativeMode(true)
	end
end

function love.mousereleased(x, y, button)
	if button == 1 then
		dragging = false
		love.mouse.setRelativeMode(false)
	end
end

function love.mousemoved(x, y, dx, dy)
	if dragging then
		angle.x = angle.x + math.rad(dx * 0.5)
		angle.y = angle.y + math.rad(dy * 0.5)
	end
end

function drawModel(model)
	love.graphics.push("all")
	--centers the world on the screen
	local w, h = love.graphics.getDimensions()
	love.graphics.translate(w/2, h/2)

	--Sends the view and model transforms to shader
	local view = matrix()
	shader:send("view", view)
	shader:send("model_matrix", makeProper(model.transform))

	--Sets the shader
	love.graphics.setShader(shader)

	--Draw mesh and clear shader.
	love.graphics.draw(model.mesh)
	love.graphics.setShader()
	love.graphics.pop()
end

